import pytest

from django.urls import reverse

from ..models import ContactMessage

valid_users = (
    ('admin', 'admin'),
    ('manager', 'manager'),
    ('user', 'user'),
)

invalid_users = (
    ('anonymous', None),
)

users = valid_users + invalid_users

status_map = {
    'messages': {
        'admin': 200, 'manager': 200, 'user': 403, 'anonymous': 302
    }
}


@pytest.mark.parametrize(('username', 'password'), users)
def test_messages(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:messages')
    response = client.get(url)
    assert response.status_code == status_map['messages'][username]


@pytest.mark.parametrize(('username', 'password'), valid_users)
def test_contact_get(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:contact')
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.parametrize(('username', 'password'), invalid_users)
def test_contact_get_anon(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:contact')
    response = client.get(url)
    assert response.status_code == 302


@pytest.mark.parametrize(('username', 'password'), valid_users)
def test_contact_post(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:contact')
    response = client.post(url, {
        'subject': 'Test',
        'message': 'This is a test.'
    })
    assert response.status_code == 200
    assert ContactMessage.objects.count() == 3


@pytest.mark.parametrize(('username', 'password'), valid_users)
def test_contact_post_invalid(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:contact')
    response = client.post(url, {})
    assert response.status_code == 200
    assert ContactMessage.objects.count() == 2


@pytest.mark.parametrize(('username', 'password'), valid_users)
def test_contact_post_cancel(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('contact:contact')
    response = client.post(url, {
        'cancel': True
    })
    assert response.status_code == 200
    assert ContactMessage.objects.count() == 2



import csv
import logging
import re
import subprocess

from django.apps import apps
from django.conf import settings

from daiquiri.core.generators import generate_csv, generate_fits, generate_votable
from daiquiri.core.utils import get_doi_url

logger = logging.getLogger(__name__)


class BaseDownloadAdapter:

    def __init__(self, database_key, database_config):
        self.database_key = database_key
        self.database_config = database_config

    def generate(self, format_key, columns, sources=[], schema_name=None, table_name=None, nrows=None,
                 query_status=None, query=None, query_language=None):
        # create the final list of arguments subprocess.Popen
        if format_key == 'sql':
            # create the final list of arguments subprocess.Popen
            self.set_args(schema_name, table_name)

            return self.generate_dump()
        else:
            # create the final list of arguments subprocess.Popen
            self.set_args(schema_name, table_name, data_only=True)

            # prepend strings with settings.FILES_BASE_PATH if they refer to files
            prepend = self.get_prepend(columns)

            if format_key == 'csv':
                return generate_csv(self.generate_rows(prepend=prepend), columns)

            elif format_key == 'votable':
                return generate_votable(self.generate_rows(prepend=prepend), columns,
                                        table=self.get_table_name(schema_name, table_name),
                                        infos=self.get_infos(query_status, query, query_language, sources),
                                        links=self.get_links(sources),
                                        services=self.get_services(),
                                        empty=(nrows==0))

            elif format_key == 'fits':
                return generate_fits(self.generate_rows(prepend=prepend), columns, nrows,
                                     table_name=self.get_table_name(schema_name, table_name))

            else:
                raise Exception('Not supported.')

    def generate_dump(self):
        # log the arguments
        logger.debug('execute "%s"', ' '.join(self.args))

        # execute the subprocess
        try:
            process = subprocess.Popen(self.args, stdout=subprocess.PIPE)

            for line in process.stdout:
                if not line.startswith((b'\n', b'\r\n', b'--', b'SET', b'/*!')):
                    yield line.decode()

        except subprocess.CalledProcessError as e:
            logger.error('Command PIPE returned non-zero exit status: %s', e)

    def generate_rows(self, prepend=None):
        # log the arguments
        logger.debug('execute "%s"', ' '.join(self.args))

        # execute the subprocess
        try:
            process = subprocess.Popen(self.args, stdout=subprocess.PIPE)

            for line in process.stdout:
                insert_pattern = re.compile('^INSERT INTO .*? VALUES \\((.*?)\\);')
                insert_result = insert_pattern.match(line.decode())
                if insert_result:
                    line = insert_result.group(1)
                    reader = csv.reader([line], quotechar="'", skipinitialspace=True)
                    row = next(reader)

                    if prepend:
                        yield [
                            (prepend[i] + cell if (i in prepend and cell != 'NULL') else cell)
                            for i, cell in enumerate(row)
                        ]
                    else:
                        yield row

        except subprocess.CalledProcessError as e:
            logger.error('Command PIPE returned non-zero exit status: %s' , e)

    def get_prepend(self, columns):
        if not settings.FILES_BASE_URL:
            return {}

        # prepend strings with settings.FILES_BASE_PATH if they refer to files
        prepend = {}

        for i, column in enumerate(columns):
            column_ucd = column.get('ucd')
            if column_ucd and 'meta.ref' in column_ucd and \
                    ('meta.file' in column_ucd or
                     'meta.note' in column_ucd or
                     'meta.image' in column_ucd):
                prepend[i] = settings.FILES_BASE_URL

        return prepend

    def get_table_name(self, schema_name, table_name):
        return f'{schema_name}.{table_name}'

    def get_infos(self, query_status, query, query_language, sources):
        infos = [
            ('QUERY_STATUS', query_status),
            ('QUERY', query),
            ('QUERY_LANGUAGE', query_language)
        ]

        for source in sources:
            infos.append(('SOURCE', '{schema_name}.{table_name}'.format(**source)))

        return infos

    def get_links(self, sources):
        return [(
            '{schema_name}.{table_name}'.format(**source),
            'doc',
            get_doi_url(source['doi']) if source['doi'] else source['url']
        ) for source in sources]

    def get_services(self):
        services = []
        if apps.is_installed('daiquiri.datalink'):
            from daiquiri.datalink.vo import get_service
            services.append(get_service())
        return services



from rest_framework import serializers

from daiquiri.metadata.models import Column


class ColumnSerializer(serializers.ModelSerializer):

    width = serializers.IntegerField(source='get_width', read_only=True)

    class Meta:
        model = Column
        fields = (
            'id',
            'order',
            'name',
            'description',
            'unit',
            'ucd',
            'datatype',
            'arraysize',
            'principal',
            'indexed',
            'std',
            'width'
        )



from daiquiri.core.renderers import ErrorRenderer, XMLRenderer

from .utils import get_job_url


class UWSRenderer(XMLRenderer):

    media_type = '*/*'

    root_attrs = {
        'xmlns:uws': 'http://www.ivoa.net/xml/UWS/v1.0',
        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xmlns:xlink': 'http://www.w3.org/1999/xlink',
        'version': '1.1'
    }

    def render_document(self, data, accepted_media_type=None, renderer_context=None):
        request = renderer_context['request']

        if isinstance(data, dict):
            if list(data.keys()) == ['results']:
                self.render_results(data['results'], request, root=True)
            elif list(data.keys()) == ['parameters']:
                self.render_parameters(data['parameters'], request, root=True)
            else:
                self.render_job(data, request)
        else:
            self.render_jobs(data, request)

    def render_jobs(self, data, request):
        self.start('uws:jobs', self.root_attrs)

        for item in data:
            self.start('uws:jobref', {
                'id': item['id'],
                'xlink:href': get_job_url(request, kwargs={'pk': item['id']}),
                'xlink:type': 'simple'
            })
            self.node('uws:phase', {}, item['phase'])
            self.node('uws:runId', {}, item['run_id'])
            self.node('uws:creationTime', {}, item['creation_time'])
            self.end('uws:jobref')

        self.end('uws:jobs')

    def render_job(self, data, request):
        self.start('uws:job', self.root_attrs)

        for key, value in data.items():
            if key == 'results':
                self.render_results(value, request)

            elif key == 'parameters':
                self.render_parameters(value, request)

            else:
                self.node('uws:' + self._to_camel_case(key), {}, value)

        self.end('uws:job')

    def render_results(self, data, request, root=False):
        self.start('uws:results', self.root_attrs if root else {})

        for result in data:
            self.start('uws:result', {
                'id': result['result_type'],
                'xlink:href': result['href'],
                'xlink:type': result['result_type']
            })
            self.end('uws:result')

        self.end('uws:results')

    def render_parameters(self, data, request, root=False):
        self.start('uws:parameters', self.root_attrs if root else {})

        for parameter_key, parameter_value in data.items():
            self.node('uws:parameter', {'id': parameter_key}, parameter_value)

        self.end('uws:parameters')


class UWSErrorRenderer(ErrorRenderer):

    media_type = '*/*'

    def render_votable(self, data, accepted_media_type=None, renderer_context=None):
        self.start('RESOURCE', {
            'type': 'results'
        })
        self.node('INFO', {
            'name': 'QUERY_STATUS',
            'value': 'ERROR'
        }, self.get_error_string(data))
        self.end('RESOURCE')



import imp

from django.core.management.base import BaseCommand


class Command(BaseCommand):

    requires_system_checks = []
    can_import_settings = False

    def handle(self, *args, **options):
        print(imp.find_module('daiquiri')[1])



from django.urls import include, path, re_path

from rest_framework import routers

from .views import reference, table
from .viewsets import ColumnViewSet, RowViewSet

app_name = 'serve'

router = routers.DefaultRouter()
router.register(r'rows', RowViewSet, basename='row')
router.register(r'columns', ColumnViewSet, basename='column')

urlpatterns = [
    re_path(r'^table/(?P<schema_name>[A-Za-z0-9_]+)/(?P<table_name>[A-Za-z0-9_]+)/$', table, name='table'),
    re_path(r'^references/(?P<key>[-\w]+)/(?P<value>.+)/$', reference, name='reference'),

    # rest api
    path('api/', include(router.urls)),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-05-15 15:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_auth', '0009_delete_detailkey'),
    ]

    operations = [
        migrations.AddField(
            model_name='profile',
            name='consent',
            field=models.BooleanField(default=False, help_text='Designates whether the user has agreed to the terms of use.', verbose_name='Consent'),
        ),
    ]



import logging
import os
import re

from django.conf import settings
from django.shortcuts import Http404
from django.utils.html import strip_tags
from django.utils.text import Truncator

from lunr import lunr

from .utils import get_file_path, read_file_content

logger = logging.getLogger(__name__)


class Searcher:

    cms_files = {}

    lunr_index = None


    @classmethod
    def search_for_string(cls, string_query):

        cls.build_lunr_index()

        try:
            lunr_results = cls.lunr_index.search(string_query)
            file_urls = {f["ref"]: f["score"] for f in lunr_results}
            match_data = {f["ref"]: f["match_data"].metadata for f in lunr_results}
            results = [f for _, f in cls.cms_files.items() if f["url"] in file_urls]
        except:  # noqa: E722
            results = []

        for res in results:
            res["score"] = file_urls[res["url"]]
            res["match_data"] = match_data[res["url"]]
        results = sorted(results, key=lambda res: -res['score'])

        return cls._reformat(results.copy())


    @classmethod
    def build_lunr_index(cls):

        if settings.FILES_DOCS_REL_PATH is None:
            logger.error('FILES_DOCS_REL_PATH is not defined')
            raise Http404

        docs_path = os.path.join(settings.FILES_BASE_PATH, settings.FILES_DOCS_REL_PATH)

        unique_files = set()
        for dir_path, _, files in os.walk(docs_path):
            for file in files:
                file_path = get_file_path(os.path.join(dir_path, file))
                if file_path and (file_path.endswith(".html") or file_path.endswith(".md")):
                    unique_files.add(file_path)


        if len(unique_files) == 0:
            logger.error('No files found in %s', docs_path)
            raise Http404


        any_changes = False

        # clean up files that do not exist anymore
        previous_cms_files = Searcher.cms_files
        for file_path in previous_cms_files:
            if file_path not in unique_files:
                any_changes = True
                cls.cms_files.pop(file_path)


        # read content of a file if it was modified since the last read
        for file_path in unique_files:
            current_mtime = os.path.getmtime(file_path)
            previous_mtime = Searcher.cms_files.get(file_path, {}).get("mtime", 0.0)
            if current_mtime != previous_mtime:
                any_changes = True
                body = read_file_content(file_path)
                title = cls.find_title_in_content(body)
                cls.cms_files[file_path] = {}
                cls.cms_files[file_path]["mtime"] = current_mtime
                cls.cms_files[file_path]["body"] = strip_tags(body)
                cls.cms_files[file_path]["title"] = strip_tags(title)
                cls.cms_files[file_path]["url"] = \
                        os.path.join(settings.FILES_BASE_URL,
                                     os.path.relpath(os.path.splitext(file_path)[0],
                                                     settings.FILES_BASE_PATH), "")

        # make an update of the lunr index if there were any changes in the files
        # since the previous search
        if any_changes:
            docs = [{
                "path": path,
                "url": doc["url"],
                "body": doc["body"],
                "title": doc["title"]
                } for path, doc in cls.cms_files.items()]
            cls.lunr_index = lunr(ref="url", fields=("body", "title",), documents = docs)



    @classmethod
    def find_title_in_content(cls, html):
        start_pattern = "<h\\d{1,5}>"
        end_pattern = "</h\\d{1,5}>"

        title = "..."
        match_start = re.search(start_pattern, html)
        if match_start:
            title_starts_at = match_start.span()[1]
            match_end = re.search(end_pattern, html[title_starts_at:])
            if match_end:
                title_ends_at = title_starts_at + match_end.span()[0]
                title = html[title_starts_at:title_ends_at]
        return title


    @classmethod
    def _reformat(cls, results):
        """ Truncates the text in the body around the search string -50:+250 chars

        """
        num_chars_before = 50
        num_chars_after = 250
        num_chars = num_chars_before + num_chars_after

        for result in results:
            result["body"] = result["body"].strip(result["title"])
            result["match_data"] = next(iter(result["match_data"].keys()))
            truncator = Truncator(result["body"])
            if result["match_data"] in result["title"]:
                result["body"] = truncator.chars(num_chars)
            else:
                search_string_pos = result["body"].find(result["match_data"])
                if search_string_pos >= 0:
                    result["body"] = truncator.chars(search_string_pos + num_chars_after)
                    if len(result["body"]) > num_chars:
                        result["body"] = "".join([
                                    "...",
                                    result["body"][-num_chars:]
                                    ])
                else:
                    result["body"] = truncator.chars(num_chars)

        return results



from django.db import models


class Schema(models.Model):

    schema_name = models.CharField(max_length=256)
    utype = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    description = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001

    class Meta:
        db_table = 'schemas'

    def __str__(self):
        return self.schema_name


class Table(models.Model):

    schema = models.ForeignKey(Schema, related_name='tables', on_delete=models.CASCADE)

    schema_name = models.CharField(max_length=256)
    table_name = models.CharField(max_length=256)
    table_type = models.CharField(max_length=256)
    utype = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    description = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    table_index = models.IntegerField(null=True, blank=True)

    class Meta:
        db_table = 'tables'

    def __str__(self):
        return f'{self.schema_name}.{self.table_name}'


class Column(models.Model):

    table = models.ForeignKey(Table, related_name='columns', on_delete=models.CASCADE)

    table_name = models.CharField(max_length=256)
    column_name = models.CharField(max_length=256)
    datatype = models.CharField(max_length=256)
    arraysize = models.IntegerField(null=True, blank=True)
    size = models.IntegerField(null=True, blank=True)
    description = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    utype = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    unit = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    ucd = models.CharField(max_length=256, null=True, blank=True)  # noqa: DJ001
    principal = models.BooleanField(default=False)
    indexed = models.BooleanField(default=False)
    std = models.BooleanField(default=False)
    column_index = models.IntegerField(null=True, blank=True)

    class Meta:
        db_table = 'columns'

    def __str__(self):
        return f'{self.table_name}.{self.column_name}'



from django.contrib.auth.models import User

from daiquiri.query.process import check_permissions


def test_columns(db):
    user = User.objects.get(username='admin')
    keywords = []
    tables = [('daiquiri_data_obs', 'stars')]
    columns = [
        ('daiquiri_data_obs', 'stars', 'ra'),
        ('daiquiri_data_obs', 'stars', 'dec')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == []


def test_columns_not_found(db):
    user = User.objects.get(username='admin')
    keywords = []
    tables = [('daiquiri_data_obs', 'stars')]
    columns = [
        ('daiquiri_data_obs', 'stars', 'ra'),
        ('daiquiri_data_obs', 'stars', 'not_found')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == ['Column not_found not found.']


def test_alias(db):
    user = User.objects.get(username='admin')
    keywords = []
    tables = [('daiquiri_data_obs', 'stars')]
    columns = [
        ('daiquiri_data_obs', 'stars', 'ra'),
        ('daiquiri_data_obs', 'stars', 'dec'),
        (None, None, 'alias')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == []


def test_missing_schema(db):
    user = User.objects.get(username='admin')
    keywords = []
    tables = [(None, 'stars')]
    columns = [
        ('daiquiri_data_obs', 'stars', 'ra'),
        ('daiquiri_data_obs', 'stars', 'dec'),
        (None, None, 'alias')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == ['No schema given for table stars.']


def test_missing_table(db):
    user = User.objects.get(username='admin')
    keywords = []
    tables = [('daiquiri_data_obs', None)]
    columns = [
        ('daiquiri_data_obs', 'stars', 'ra'),
        ('daiquiri_data_obs', 'stars', 'dec'),
        (None, None, 'alias')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == ['No table given for schema daiquiri_data_obs.']


def test_user_schema(db):
    user = User.objects.get(username='user')
    keywords = []
    tables = [('daiquiri_user_user', 'test')]
    columns = [
        (None, None, '*')
    ]
    functions = []

    result = check_permissions(user, keywords, tables, columns, functions)

    assert result == []



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-02-14 13:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_metadata', '0003_groups'),
    ]

    operations = [
        migrations.AddField(
            model_name='function',
            name='query_string',
            field=models.CharField(default='', max_length=256),
            preserve_default=False,
        ),
    ]



from django.contrib import admin

from .models import Record


class RecordAdmin(admin.ModelAdmin):
    search_fields = ('resource', 'user__username')
    list_display = ('time', 'resource_type', 'client_ip', 'user')
    list_display_links = ('time', )


admin.site.register(Record, RecordAdmin)



from rest_framework import serializers

from daiquiri.core.serializers import CaseInsensitiveSerializer
from daiquiri.jobs.models import Job

from .utils import get_job_results


class JobListSerializer(serializers.ModelSerializer):

    creation_time = serializers.SerializerMethodField()

    class Meta:
        model = Job
        fields = ('id', 'phase', 'run_id', 'creation_time')

    def get_creation_time(self, obj):
        return obj.creation_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if obj.creation_time else None


class JobRetrieveSerializer(serializers.ModelSerializer):

    job_id = serializers.UUIDField(source='id')
    owner_id = serializers.SerializerMethodField()
    destruction = serializers.DateTimeField(source='destruction_time')
    results = serializers.SerializerMethodField()

    creation_time = serializers.SerializerMethodField()
    start_time = serializers.SerializerMethodField()
    end_time = serializers.SerializerMethodField()

    class Meta:
        model = Job
        fields = (
            'job_id',
            'run_id',
            'owner_id',
            'phase',
            'quote',
            'creation_time',
            'start_time',
            'end_time',
            'execution_duration',
            'destruction',
            'error_summary',
            'results',
            'parameters'
        )

    def get_owner_id(self, obj):
        if obj.owner:
            return obj.owner.username
        else:
            return None

    def get_results(self, obj):
        return get_job_results(self.context['request'], obj)

    def get_creation_time(self, obj):
        return obj.creation_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if obj.creation_time else None

    def get_start_time(self, obj):
        return obj.start_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if obj.start_time else None

    def get_end_time(self, obj):
        return obj.end_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if obj.end_time else None


class JobUpdateSerializer(CaseInsensitiveSerializer):

    ACTION = serializers.CharField(required=False)
    DESTRUCTION = serializers.DateTimeField(required=False)
    EXECUTIONDURATION = serializers.IntegerField(required=False)
    PHASE = serializers.CharField(required=False)


class SyncJobSerializer(CaseInsensitiveSerializer):

    RESPONSEFORMAT = serializers.CharField(required=False)
    MAXREC = serializers.IntegerField(required=False)
    RUNID = serializers.CharField(required=False, default='')


class AsyncJobSerializer(CaseInsensitiveSerializer):

    PHASE = serializers.CharField(required=False)
    RESPONSEFORMAT = serializers.CharField(required=False)
    MAXREC = serializers.IntegerField(required=False)
    RUNID = serializers.CharField(required=False, default='')



from django.urls import include, path

from rest_framework import routers

from .views import ManagementView
from .viewsets import RecordViewSet

app_name = 'stats'

router = routers. DefaultRouter()
router.register(r'records', RecordViewSet, basename='record')

urlpatterns = [
    path('api/', include(router.urls)),

    path('management/', ManagementView.as_view(), name='management'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-02-17 19:41
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_auth', '0008_profile_is_pending'),
    ]

    operations = [
        migrations.DeleteModel(
            name='DetailKey',
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-22 09:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_stats', '0008_alter_record_resource'),
    ]

    operations = [
        migrations.AddField(
            model_name='record',
            name='size',
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
    ]



import daiquiri.core.env as env

ASYNC = env.get_bool('ASYNC')

QUEUES = [
    {
        'key': 'default',
        'concurrency': 1
    },
    {
        'key': 'download',
        'concurrency': 1
    }
]

IPV4_PRIVACY_MASK = 16
IPV6_PRIVACY_MASK = 32

SITE_URL = env.get('SITE_URL')

SITE_IDENTIFIER = None
SITE_TITLE = None
SITE_DESCRIPTION = None
SITE_LICENSE = None
SITE_CREATOR = None
SITE_CONTACT = None
SITE_PUBLISHER = None
SITE_CREATED = None
SITE_UPDATED = None
SITE_LANGUAGE = 'en'
SITE_SUBJECTS = [
    {
        'subject': 'Astronomy',
        'subjectScheme': 'Library of Congress Subject Headings (LCSH)',
        'schemeURI': 'http://id.loc.gov/authorities/subjects',
        'valueURI': 'http://id.loc.gov/authorities/subjects/sh85009003'
    }
]
SITE_TYPE = 'service'
SITE_LOGO_URL = None



import csv
import datetime
import io
import struct
import sys
from xml.sax.saxutils import escape, quoteattr

from django.contrib.sites.models import Site

from daiquiri import __version__ as daiquiri_version


def generate_csv(generator, fields):
    if sys.version_info.major >= 3:
        io_class = io.StringIO
    else:
        io_class = io.BytesIO

    # write header
    f = io_class()
    csv.writer(f, quotechar='"').writerow([field['name'] for field in fields])
    yield f.getvalue()

    for row in generator:
        if row:
            # convert curl brace to square brace in all array-like columns
            corrected_row = []
            for col in row:

                corrected_col = col

                if isinstance(col, str):
                    if col.startswith('{') and col.endswith('}'):
                        corrected_col = col.replace('{', '[').replace('}', ']')

                corrected_row = [*corrected_row, corrected_col]

            f = io_class()
            csv.writer(f, quotechar='"').writerow(corrected_row)
            yield f.getvalue()


def correct_col_for_votable(col):
    corrected_col = col

    if col.startswith('{') and col.endswith('}'): # this is an array
        # remove {} and replace , with space
        corrected_col = col.replace('{', '').replace('}', '').replace(',', ' ')

    return corrected_col

def generate_votable(generator, fields, infos=[], links=[], services=[], table=None, empty=None):
    yield '''<?xml version="1.0"?>
<VOTABLE version="1.3"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns="http://www.ivoa.net/xml/VOTable/v1.3"
    xmlns:stc="http://www.ivoa.net/xml/STC/v1.30">'''

    yield '''
    <RESOURCE type="results">'''

    for key, value in infos:
        if value is not None:
            yield f'''
            <INFO name={quoteattr(key)} value={quoteattr(value)} />'''

    for title, content_role, href in links:
        yield f'''
        <LINK title={quoteattr(title)} content-role={quoteattr(content_role)} href={quoteattr(href)}/>'''

    if table is not None:
        yield f'''
        <TABLE name="{table}">'''
    else:
        yield '''
        <TABLE>'''

    for field in fields:
        attrs = []
        for key in ['name', 'unit', 'ucd', 'utype']:
            if field.get(key):
                value = field[key].replace('&', '&amp;') \
                                  .replace('"', '&quot;') \
                                  .replace("'", '&apos;') \
                                  .replace('<', '&lt;') \
                                  .replace('>', '&gt;')
                attrs.append(f'{key}="{value}"')

        if field.get('ucd'):
            if 'meta.id' in field['ucd'] and 'meta.ref' in field['ucd']:
                attrs.append('ID="datalinkID"')

        if 'arraysize' in field:
            if field.get('datatype') == 'char' and field['arraysize'] is None:
                attrs.append('arraysize="*"')
            elif field['arraysize']:
                attrs.append('arraysize="{}"'.format(field['arraysize']))

        if 'datatype' in field:
            if field['datatype'] in ['boolean', 'char', 'unsignedByte', 'short', 'int', 'long', 'float', 'double']:
                attrs.append('datatype="{}"'.format(field['datatype']))
            else:
                attrs.append('xtype="{}"'.format(field['datatype']))

        if attrs:
            yield '''
            <FIELD {} />'''.format(' '.join(attrs))

    if not empty:
        yield '''
            <DATA>
                <TABLEDATA>'''

        # write rows of the table yielded by the generator
        for row in generator:
            yield '''
                    <TR>
                        <TD>{}</TD>
                    </TR>'''.format('''</TD>
                        <TD>'''.join([
                            ('' if cell in ['NULL', None] else correct_col_for_votable(escape(str(cell))))
                            for cell in row
                        ]))

        yield '''
                </TABLEDATA>
            </DATA>'''
    yield '''
        </TABLE>
    </RESOURCE>'''

    for service in services:
        yield '''
    <RESOURCE type="meta" utype="adhoc:service">'''
        for param in service.get('params', []):
            yield '''
        <PARAM name="{name}" datatype="{datatype}" arraysize="{arraysize}" value="{value}" />'''.format(**param)

        for group in service.get('groups', []):
            yield '''
        <GROUP name="{name}">'''.format(**group)
            for param in group.get('params', []):
                yield '''
            <PARAM name="{name}" datatype="{datatype}" arraysize="{arraysize}" value="{value}" ref="{ref}"/>'''.format(**param)  # noqa: E501
            yield '''
        </GROUP>'''
        yield '''
    </RESOURCE>'''
    yield '''
</VOTABLE>
'''


def generate_fits(generator, fields, nrows, table_name=None):

    # VO format label, FITS format label, size, NULL value, encoded value
    formats_dict = {
        'boolean':      ('s', 'L', 1,  b'\x00',             lambda x: b'T' if x == 'true' else b'F'),
        'short':        ('h', 'I', 2,  32767,               int),
        'int':          ('i', 'J', 4,  2147483647,          int),
        'long':         ('q', 'K', 8,  9223372036854775807, int),
        'float':        ('f', 'E', 4,  float('nan'),        float),
        'double':       ('d', 'D', 8,  float('nan'),        float),
        'char':         ('s', 'A', 32, b'',                 lambda x: x.encode()),
        'timestamp':    ('s', 'A', 19, b'',                 lambda x: x.encode()),
        'array':        ('s', 'A', 64, b'',                 lambda x: x.encode()),
        'spoint':       ('s', 'A', 64, b'',                 lambda x: x.encode()),
        'unknown':      ('s', 'A', 8,  b'',                 lambda x: x.encode())
    }

    names = [field['name'] for field in fields]
    datatypes = [field['datatype'] for field in fields]
    arraysizes = [field.get('arraysize') or '' for field in fields]

    for i, d in enumerate(zip(datatypes, arraysizes)):
        if d[0] == 'timestamp':
            arraysizes[i] = formats_dict['timestamp'][2]
        elif d[0] in ('char', 'spoint', 'array') and d[1] == '':
            arraysizes[i] = formats_dict[d[0]][2]
        elif d[0] is None:
            datatypes[i] = 'unknown'
            arraysizes[i] = formats_dict['unknown'][2]

    units = []
    ucds = []
    for d in fields:
        if 'unit' in d and d['unit'] is not None:
            units.append(d['unit'])
        else:
            units.append('')
        if 'ucd' in d and d['ucd'] is not None:
            ucds.append(d['ucd'])
        else:
            ucds.append('')

    naxis1 = sum([formats_dict[i[0]][2] if not i[1] else i[1]
                  for i in zip(datatypes, arraysizes)])
    naxis2 = nrows
    tfields = len(names)

    site = str(Site.objects.get_current())[:30]

    content = """

                                  `,......`
                               :::.````````...
                             ::``,:::,`.....``..`
                           ,:.`,``,:::`....`````..
                          .:``:::,``,:`::```...``..
                          :,`::::::,`````,::::::`.:
                          :::::::::::::::::::::::::
         '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
         ,:,,,,,,,,,,,,,,,,,,,,,,,,,,,,,............................
           :,,,`````````...,,,,,,,,............................```
             ,,,.```````...,,,,,,,,............................`
              `,,,``````....,,,,,,,...........................`
                :+++,,,,,::::+++++++++++++++++++++++++++++++'
                  \\\\     created with django-daiquiri     //
                    \\\\               v%s//
                      \\\\%s//
                        ,,,.....,,,..................`
                         :++':;;;++++++++++++++++++'
                           ,,,,,,,::::;;;;;;;;''''
                             ,,,................
                               ,,,............
                                 ,,.........`
                                  `;;;;;;;:
                                   ,'` `''
                                    :: ::
                                     . .`
        +++++++++:             ,                       ,          .,
         +++    +++           '+'                     '+'         ++.
         +++    ,++:  :'+'.  `'+'   ,'';'+ ;'+` .'+' `'+' .++,:+ ;++`
         +++    `++; ++  '';  ;'; `''  `''  ''`  ;''  ;''  ;++:,  ++`
         +++    ;++. :++''''  ;'; ;''  `''  ''`  ;''  ;''  ;+'    ++`
         +++  `+++: .++  '''  ;'' .''  .''  '',  '''  ;''  '+'    ++`
        +++++++:     `'+'`+',,+++,  ;+'.++   ;++: +'.,+++,,++++. ++++
                                      `+++.

    """.replace('\x0a', ' ') % (daiquiri_version.ljust(18),
           (' ' * (15 - len(site) // 2) + site).ljust(30))

    # Main header #############################################################
    header0 = [i.ljust(80) for i in [
        'SIMPLE  =                    T / conforms to FITS standard',
        'BITPIX  =                    8 / array data type',
        'NAXIS   =                    1 / number of array dimensions',
        'NAXIS1  =                 2880 / number of characters',
        'EXTEND  =                    T',
        'NTABLE  =                    1',
        'END'
        ]]

    h0 = ''.join(header0)
    h0 += ' ' * (2880 * (len(h0) // 2880 + 1) - len(h0))

    yield h0.encode()

    # Main table content - required by some FITS viewers ######################

    yield (content + '\x00' * (2880 - len(content))).encode()

    # Table header ############################################################
    header1 = [
       "XTENSION= 'BINTABLE'           / binary table extension".ljust(80),
       'BITPIX  =                    8 / array data type'.ljust(80),
       'NAXIS   =                    2 / number of array dimensions'.ljust(80),
       'NAXIS1  = %20d / length of dimension 1'.ljust(64),
       'NAXIS2  = %20d / length of dimension 2'.ljust(64),
       'PCOUNT  =                    0 / number of group parameters'.ljust(80),
       'GCOUNT  =                    1 / number of groups'.ljust(80),
       'TFIELDS = %20d / number of table fields'.ljust(64),
        ]
    if table_name is not None:
        # table_name needs to be shorter than 68 chars
        header1.append((f"EXTNAME = '{table_name[:68]!s}' / table name").ljust(80))

    h1 = ''.join(header1) % (naxis1, naxis2, tfields)

    ttype = ("TTYPE%s", "= '%s'")
    tform = ("TFORM%s", "= '%s'")
    tnull = ("TNULL%s", "=  %s")
    tunit = ("TUNIT%s", "= '%s'")
    tucd = ("TUCD%s", "= '%s'")

    for i, d in enumerate(zip(names, datatypes, arraysizes, units, ucds)):
        temp = "".join(((ttype[0] % str(i + 1)).ljust(8),
                       ttype[1] % d[0][:68].ljust(8)))[:80].ljust(30)
        temp += ' / label for column %d' % (i + 1)
        temp = temp[:80]
        temp += ' ' * (80 - len(temp))
        h1 += temp

        ff = (str(d[2]) + formats_dict[d[1]][1]).ljust(8)
        temp = "".join(((tform[0] % str(i + 1)).ljust(8),
                       tform[1] % ff))[:80].ljust(31)
        temp += '/ format for column %d' % (i + 1)
        temp = temp[:80]
        temp += ' ' * (80 - len(temp))
        h1 += temp

        # NULL values only for int-like types
        if d[1] in ('short', 'int', 'long'):
            temp = "".join(((tnull[0] % str(i + 1)).ljust(8),
                           tnull[1] % formats_dict[d[1]][3]))[:80].ljust(31)
            temp += '/ blank value for column %d' % (i + 1)
            temp = temp[:80]
            temp += ' ' * (80 - len(temp))
            h1 += temp

        if d[3]:
            temp = "".join(((tunit[0] % str(i + 1)).ljust(8),
                           tunit[1] % d[3][:68].ljust(8)))[:80].ljust(30)
            temp += ' / unit for column %d' % (i + 1)
            temp = temp[:80]
            temp += ' ' * (80 - len(temp))
            h1 += temp

        if d[4]:
            temp = "".join(((tucd[0] % str(i + 1)).ljust(8),
                           tucd[1] % d[4][:68].ljust(8)))[:80].ljust(30)
            temp += ' / ucd for column %d' % (i + 1)
            temp = temp[:80]
            temp += ' ' * (80 - len(temp))
            h1 += temp

    now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
    h1 += (f"DATE-HDU= '{now}' / UTC date of HDU creation").ljust(80)

    h1 += 'END'.ljust(80)
    h1 += ' ' * (2880 * (len(h1) // 2880 + 1) - len(h1))

    yield h1.encode()

    # Data ####################################################################
    fmt = '>' + ''.join([str(i[1]) + formats_dict[i[0]][0]
                         for i in zip(datatypes, arraysizes)])

    row_count = 0
    for row in generator:
        r = [formats_dict[i[1]][3] if i[0] == 'NULL'
             else formats_dict[i[1]][4](i[0]) for i in zip(row, datatypes)]

        yield struct.pack(fmt, *r)
        row_count += 1

    # Footer padding (to fill the last block to 2880 bytes) ###################
    ln = naxis1 * row_count
    footer = '\x00' * (2880 * (ln // 2880 + 1) - ln)

    yield footer.encode()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-14 12:47
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_metadata', '0012_remove_directory_metadata_access_level'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='directory',
            name='groups',
        ),
        migrations.DeleteModel(
            name='Directory',
        ),
    ]



import re
from datetime import datetime, timedelta

from django.contrib.auth.models import User
from django.core.management.base import BaseCommand, CommandError

from daiquiri.query.models import QueryJob


class Command(BaseCommand):

    pattern = re.compile(r'(\d+)([dsmhw])')

    def add_arguments(self, parser):
        parser.add_argument('username', help='Remove all query jobs of this user.')
        parser.add_argument('-k', '--keep', default=None, help='Keep jobs newer than this timespan.')
        parser.add_argument('-y', '--yes', action='store_true', help='Answer yes for all questions.')

    def handle(self, *args, **options):

        # parse username -> owner
        if options['username'] == 'anonymous':
            owner = None
        else:
            try:
                owner = User.objects.get(username=options['username'])
            except User.DoesNotExist as e:
                raise CommandError('User "{}" does not exist'.format(options['username'])) from e

        # parse keep -> before
        if options['keep']:
            match = self.pattern.match(options['keep'])
            if match:
                value, attribute = int(match.group(1)), match.group(2)

                if attribute == 'd':
                    delta = timedelta(days=value)
                elif attribute == 's':
                    delta = timedelta(seconds=value)
                elif attribute == 'm':
                    delta = timedelta(minutes=value)
                elif attribute == 'h':
                    delta = timedelta(hours=value)
                elif attribute == 'w':
                    delta = timedelta(weeks=value)

                before = datetime.now() - delta

            else:
                raise CommandError('User "{}" does not exist'.format(options['username']))
        else:
            before = None

        # get the jobs queryset
        jobs = QueryJob.objects.filter_by_owner(owner).exclude(phase=QueryJob.PHASE_ARCHIVED)

        if before:
            jobs = jobs.exclude(creation_time__gt=before)

        count = jobs.count()
        if count > 0:
            if options['yes']:
                archive = True
            else:
                print('%i jobs found. Really archive these jobs? [yes/no]:' % jobs.count())
                archive = input().lower() in ['yes', 'y']

            if archive:
                for job in jobs:
                    job.archive()
        else:
            print('No jobs found for this user.')



from django.conf import settings
from django.urls import reverse

from daiquiri.core.vo import get_curation

from .constants import DATALINK_CONTENT_TYPE


def get_availability():
    return {
        'available': 'true',
        'note': 'service is accepting queries'
    }


def get_capabilities():
    return [
        {
            'id': 'ivo://ivoa.net/std/DataLink#links-1.0',
            'interface': {
                'type': 'vs:ParamHTTP',
                'role': 'std',
                'version': '1.0',
                'access_url': {
                    'use': 'base',
                    'url': settings.SITE_URL.rstrip('/') + reverse('datalink:link-list').rstrip('/')
                },
                'query_types': ['GET', 'POST'],
                'result_type': DATALINK_CONTENT_TYPE,
                'params': [
                    {
                        'name': 'ID',
                        'std': 'true',
                        'use': 'required',
                        'description': 'publisher dataset identifier',
                        'ucd': 'meta.id;meta.main',
                        'datatype': 'string'
                    }
                ]
            }
        }
    ]


def get_resource():
    return {
        'service': 'datalink',
        'identifier': f'ivo://{settings.SITE_IDENTIFIER}/datalink',
        'title': f'{settings.SITE_TITLE} Datalink Service',
        'curation': get_curation(),
        'content': {
            'subjects': [],
            'type': '',
            'description': f'The Datalink Service for {settings.SITE_IDENTIFIER}.',
            'referenceURL': settings.SITE_URL.rstrip('/') + reverse('datalink:root').rstrip('/')
        },
        'capabilities': get_capabilities(),
        'created': settings.SITE_CREATED,
        'updated': settings.SITE_UPDATED,
        'type': '',
        'status': 'active'
    }


def get_service():
    return {
        'params': [
            {
                'name': 'standardID',
                'datatype': 'char',
                'arraysize': '*',
                'value': 'ivo://ivoa.net/std/DataLink#links-1.0'
            },
                                {
                'name': 'accessURL',
                'datatype': 'char',
                'arraysize': '*',
                'value': settings.SITE_URL.rstrip('/') + '/datalink/links'
            }
        ],
        'groups': [
            {
                'name': 'inputParams',
                'params': [
                    {
                        'name': 'ID',
                        'datatype': 'char',
                        'arraysize': '*',
                        'value': '',
                        'ref': 'datalinkID'
                    }
                ]
            }
        ]
    }



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.utils.timezone import now
from django.views.generic import TemplateView

from daiquiri.core.views import CSRFViewMixin, ModelPermissionMixin, StoreIdViewMixin

from .forms import ContactForm
from .utils import send_contact_message


@login_required()
def contact(request):
    contact_form = ContactForm(request.POST or None)

    if request.method == 'POST':
        if 'cancel' in request.POST:
            return render(request, 'core/home.html')

        if contact_form.is_valid():

            message = contact_form.save(commit=False)
            message.set_status_active()
            message.created = now()

            message.user = request.user
            message.email = request.user.email
            message.author = request.user.profile.full_name

            message.save()

            send_contact_message(request, message)

            return render(request, 'contact/thanks.html')

    else:
        if request.user.is_authenticated:
            try:
                contact_form.initial = {
                    'email': request.user.email,
                    'author': request.user.profile.full_name
                }
            except AttributeError:
                pass

    return render(request, 'contact/contact.html', {'form': contact_form})


class MessagesView(ModelPermissionMixin, CSRFViewMixin, StoreIdViewMixin, TemplateView):
    template_name = 'contact/messages.html'
    permission_required = 'daiquiri_contact.view_contactmessage'



import pytest

from django.urls import reverse

users = (
    ('admin', 'admin'),
    ('user', 'user'),
    ('anonymous', None),
)


@pytest.mark.parametrize(('username', 'password'), users)
def test_root(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('tap:root')
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.parametrize(('username', 'password'), users)
def test_availability(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('tap:availability')
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.parametrize(('username', 'password'), users)
def test_capabilities(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('tap:capabilities')
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.parametrize(('username', 'password'), users)
def test_tables(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('tap:tables')
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.parametrize(('username', 'password'), users)
def test_examples(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse('tap:examples')
    response = client.get(url)
    assert response.status_code == 200



DATALINK_ADAPTER = 'daiquiri.datalink.adapter.DefaultDatalinkAdapter'
DATALINK_TABLES = []
DATALINK_CUSTOM_SEMANTICS = {
    "#doi": "The access_url points to the Digital Object Identifier (DOI) of the object.",
}



from django.contrib import admin
from django.forms import ChoiceField, ModelForm

from daiquiri.jobs.admin import JobAdmin

from .models import DownloadJob, Example, QueryArchiveJob, QueryJob
from .utils import get_query_download_format_choices, get_query_language_choices, get_queue_choices


class QueryJobForm(ModelForm):
    queue = ChoiceField(choices=get_queue_choices())
    query_language = ChoiceField(choices=get_query_language_choices())


class DownloadJobForm(ModelForm):
    format_key = ChoiceField(choices=get_query_download_format_choices())


class ExampleForm(ModelForm):
    query_language = ChoiceField(choices=get_query_language_choices())


@admin.register(QueryJob)
class QueryJobAdmin(JobAdmin):
    form = QueryJobForm

    search_fields = [*JobAdmin.search_fields, 'schema_name', 'table_name']
    list_display = [*JobAdmin.list_display, 'schema_name', 'table_name', 'nrows']
    list_filter = [*JobAdmin.list_filter, 'query_language', 'queue']
    actions = ['abort_job', 'archive_job']


@admin.register(DownloadJob)
class DownloadJobAdmin(JobAdmin):
    form = DownloadJobForm

    search_fields = [*JobAdmin.search_fields, 'query_job__schema_name', 'query_job__table_name', 'format_key']
    list_display = [*JobAdmin.list_display, 'query_job', 'file_path']
    list_filter = [*JobAdmin.list_filter, 'format_key']
    actions = ['abort_job', 'archive_job']


@admin.register(QueryArchiveJob)
class QueryArchiveJobAdmin(JobAdmin):
    search_fields = [*JobAdmin.search_fields, 'query_job__schema_name', 'query_job__table_name', 'column_name', 'files']
    list_display = [*JobAdmin.list_display, 'query_job', 'file_path']
    actions = ['abort_job', 'archive_job']


@admin.register(Example)
class ExampleAdmin(admin.ModelAdmin):
    form = ExampleForm

    search_fields = ('name', 'query_string')
    list_display = ('order', 'name', 'query_string')
    list_display_links = ('name', )
    list_filter = ('query_language', )



from django import forms
from django.conf import settings
from django.utils.translation import gettext as _

from daiquiri.core.forms import HoneypotField
from daiquiri.core.utils import sanitize_str

from .models import ContactMessage


class ContactForm(forms.ModelForm):

    use_required_attribute = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # add honeypot field
        if settings.HONEYPOT_ENABLED is True:
            self.fields[sanitize_str(settings.HONEYPOT_FIELD_NAME)] = HoneypotField()

    class Meta:
        model = ContactMessage
        fields = ['subject', 'message']
        widgets = {
            'subject': forms.TextInput(attrs={'placeholder': _('Subject')}),
            'message': forms.Textarea(attrs={'placeholder': _('Your message')}),
        }
        labels = {
            'subject': _('Subject'),
            'message': _('Your message')
        }



from daiquiri.core.adapter import DatabaseAdapter


def test_fetch_table(db):
    row = DatabaseAdapter().fetch_table('daiquiri_data_obs', 'stars')
    assert row['type'] == 'table'


def test_fetch_table_error(db):
    row = DatabaseAdapter().fetch_table('daiquiri_data_obs', 'stars1')
    assert row == {}


def test_fetch_row(db):
    row = DatabaseAdapter().fetch_row('daiquiri_data_obs', 'stars', column_names=None, search=None, filters={
        'id': '4551299946478123136'
    })
    assert row[0] == 4551299946478123136


def test_fetch_rows(db):
    rows = DatabaseAdapter().fetch_rows('daiquiri_data_obs', 'stars')
    assert len(rows) == 10
    assert len(rows[0]) == 4
    assert rows[0][0] == 1714709274237701248


def test_fetch_rows_filter(db):
    rows = DatabaseAdapter().fetch_rows('daiquiri_data_sim', 'halos', filters={
        'id': '85000000000'
    })
    assert len(rows) == 1
    assert len(rows[0]) == 8
    assert rows[0][0] == 85000000000


def test_fetch_nrows(db):
    nrows = DatabaseAdapter().fetch_nrows('daiquiri_data_obs', 'stars')
    assert nrows == 10000


def test_fetch_size(db):
    size = DatabaseAdapter().fetch_size('daiquiri_data_obs', 'stars')
    assert size > 0


def test_fetch_columns(db):
    columns = DatabaseAdapter().fetch_columns('daiquiri_data_obs', 'stars')
    assert len(columns) == 4
    assert columns[0]['indexed'] is True


def test_fetch_column(db):
    column = DatabaseAdapter().fetch_column('daiquiri_data_obs', 'stars', 'id')
    assert column['name'] == 'id'
    assert column['datatype']
    assert column['indexed'] is True


def test_fetch_column_names(db):
    columns = DatabaseAdapter().fetch_column_names('daiquiri_data_obs', 'stars')
    assert columns[0] == 'id'


def test_count_rows(db):
    row = DatabaseAdapter().count_rows('daiquiri_data_obs', 'stars')
    assert row == 10000



from django.contrib.auth.models import User
from django.core.management.base import BaseCommand

from daiquiri.core.adapter import DatabaseAdapter
from daiquiri.query.models import QueryJob


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--user', help='Only scrub jobs for this user.')
        parser.add_argument('--archive', action='store_true', help='Archive stale jobs.')

    def handle(self, *args, **options):
        if options['user']:
            if options['user'] == 'anonymous':
                owners = [None]
            else:
                owners = [User.objects.get(username=options['user'])]

        else:
            owners = [None, *list(User.objects.all())]

        adapter = DatabaseAdapter()

        stale_jobs = []
        for owner in owners:
            jobs = QueryJob.objects.filter(owner=owner)
            for job in jobs:
                if job.phase == job.PHASE_COMPLETED:
                    if not adapter.fetch_table(job.schema_name, job.table_name):
                        stale_jobs.append(job)

        if stale_jobs:
            print('The following QueryJobs have no associated database table:')

            for job in stale_jobs:
                username = job.owner.username if job.owner else 'anonymous'
                print(f'{job.id} by {username} -> {job.schema_name}.{job.table_name}')

            if options['archive']:
                for job in stale_jobs:
                    job.archive()

                print('The jobs have been archived.')
        else:
            print('No QueryJobs without associated associated database table have been found.')



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-07-24 15:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_jobs', '0009_meta'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='max_records',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='job',
            name='response_format',
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='job',
            name='run_id',
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
    ]



# Generated by Django 2.1.3 on 2018-11-05 16:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_query', '0020_python3'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='downloadjob',
            options={'ordering': ('start_time',), 'verbose_name': 'DownloadJob', 'verbose_name_plural': 'DownloadJobs'},
        ),
        migrations.AlterModelOptions(
            name='example',
            options={'ordering': ('order',), 'verbose_name': 'Example', 'verbose_name_plural': 'Examples'},
        ),
        migrations.AlterModelOptions(
            name='queryarchivejob',
            options={'ordering': ('start_time',), 'verbose_name': 'QueryArchiveJob', 'verbose_name_plural': 'QueryArchiveJob'},
        ),
        migrations.AlterModelOptions(
            name='queryjob',
            options={'ordering': ('start_time',), 'verbose_name': 'QueryJob', 'verbose_name_plural': 'QueryJobs'},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-10 14:18
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_jobs', '0012_meta'),
        ('daiquiri_query', '0014_downloadjob'),
    ]

    operations = [
        migrations.CreateModel(
            name='QueryArchiveJob',
            fields=[
                ('job_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='daiquiri_jobs.Job')),
                ('column_name', models.CharField(help_text='Column name for this download.', max_length=32, verbose_name='Column name')),
                ('files', jsonfield.fields.JSONField(help_text='List of files in the archive.', verbose_name='Files')),
                ('job', models.ForeignKey(help_text='QueryJob this ArchiveJob belongs to.', on_delete=django.db.models.deletion.CASCADE, related_name='archives', to='daiquiri_query.QueryJob', verbose_name='QueryJob')),
            ],
            options={
                'ordering': ('start_time',),
                'verbose_name': 'QueryArchiveJob',
                'verbose_name_plural': 'QueryArchiveJob',
                'permissions': (('view_queryarchivejob', 'Can view QueryArchiveJob'),),
            },
            bases=('daiquiri_jobs.job',),
        ),
        migrations.RemoveField(
            model_name='downloadjob',
            name='file_path',
        ),
    ]



from django.http import Http404
from django.shortcuts import redirect, render

from daiquiri.metadata.utils import get_user_columns, get_table_metadata

from .utils import get_resolver


def table(request, schema_name, table_name):

    if get_user_columns(request.user, schema_name, table_name):
        table = get_table_metadata(request.user, schema_name, table_name)
        context = {}
        if table is not None:
            context['table_title'] = table.title
            context['table_name'] = table.name
            context['schema_name'] = table.schema.name
            context['table_description'] = table.long_description
        return render(request, 'serve/table.html', context)

    # if nothing worked, return 404
    raise Http404


def reference(request, key, value):

    resolver = get_resolver()
    if resolver is None:
        raise Http404()

    url = resolver.resolve(key, value)
    if url is None:
        raise Http404()

    return redirect(url)



import pytest

from django.test import override_settings
from django.urls import reverse

users = (
    ('admin', 'admin'),
    ('user', 'user'),
    ('evil', 'evil'),
    ('anonymous', None),
)

status_map = {
    'list': {
        'admin': 200, 'user': 200, 'evil': 200, 'anonymous': 403
    }
}

urlnames = {
    'list': 'query:status-list'
}


@pytest.mark.parametrize(('username', 'password'), users)
def test_list(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse(urlnames['list'])
    response = client.get(url)
    assert response.status_code == status_map['list'][username], response.json()

    if response.status_code == 200:
        assert response.json().get('guest') is False


@override_settings(QUERY_ANONYMOUS=True)
def test_list_anonymous(db, client):
    url = reverse(urlnames['list'])
    response = client.get(url)
    assert response.status_code == 200, response.json()

    if response.status_code == 200:
        assert response.json().get('guest') is True



from __future__ import unicode_literals

from django.db import migrations


def run_data_migration(apps, schema_editor):
    QueryJob = apps.get_model('daiquiri_query', 'QueryJob')

    for queryjob in QueryJob.objects.filter(job_type='SYNC'):
        queryjob.delete()


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_query', '0018_blank_metadata'),
        ('daiquiri_stats', '0002_data_migration'),
    ]

    operations = [
        migrations.RunPython(run_data_migration)
    ]



from django import template

from daiquiri.contact.models import AnnouncementMessage

register = template.Library()


@register.inclusion_tag("contact/announcements.html", takes_context=True)
def show_announcements(context):
    request = context["request"]
    announcements = AnnouncementMessage.objects.filter(visible=True)
    announcements = [msg for msg in announcements if msg.get_filter()(request) is True]
    return {
        "announcements": announcements,
    }



from django.http import FileResponse, JsonResponse

from rest_framework import viewsets

from django_user_agents.utils import get_user_agent

from daiquiri.core.generators import generate_votable

from .adapter import DatalinkAdapter
from .constants import DATALINK_CONTENT_TYPE, DATALINK_FIELDS


class SyncDatalinkJobViewSet(viewsets.GenericViewSet):
    '''
    Generate the datalink VOTable
    '''

    def list(self, request):
        return self.perform_sync_job(request, request.GET)

    def create(self, request):
        return self.perform_sync_job(request, request.POST)

    def perform_sync_job(self, request, data):
        if 'ID' in data:
            identifiers = data.getlist('ID')
        else:
            identifiers = []

        adapter = DatalinkAdapter()

        # get all datalink entries (DatalinkTable, Metadata and Dynamic)
        rows = adapter.get_datalink_rows(identifiers)

        if data.get('RESPONSEFORMAT') == 'application/json':
            return JsonResponse({
                'links': [
                    {
                        'ID': row[0],
                        'access_url': row[1],
                        'service_def': row[2],
                        'error_message': row[3],
                        'description': row[4],
                        'semantics': row[5],
                        'content_type': row[6],
                        'content_length': row[7],
                    } for row in rows if not row[3]
                ]
            })
        else:
            datalink_table = generate_votable(rows, DATALINK_FIELDS)

            user_agent = get_user_agent(request)
            content_type = 'application/xml' if user_agent.is_pc else DATALINK_CONTENT_TYPE
            return FileResponse(datalink_table, content_type=content_type)



from django.conf import settings

from daiquiri.core.constants import ACCESS_LEVEL_INTERNAL, ACCESS_LEVEL_PUBLIC

from .models import Column as TapColumn
from .models import Schema as TapSchema
from .models import Table as TapTable


def check_tap_visibility(obj):
    # check if the metadata_access_level of the object is
    #   (a) PUBLIC or INTERNAL if QUERY_ANONYMOUS is True
    # or
    #   (b) INTERNAL if QUERY_ANONYMOUS is False
    return (obj.metadata_access_level == ACCESS_LEVEL_PUBLIC) or \
        (obj.metadata_access_level == ACCESS_LEVEL_INTERNAL and not settings.QUERY_ANONYMOUS)


def update_schema(schema):
    '''
    Update or create the schema in the TAP_SCHEMA.
    '''
    if check_tap_visibility(schema):
        tap_schema, created = TapSchema.objects.get_or_create(pk=schema.id)

        tap_schema.schema_name = schema.name
        tap_schema.utype = None
        if schema.description:
            tap_schema.description = schema.description[:255]
        tap_schema.save()

        if created:
            # call the handler for each table of the schema
            for table in schema.tables.all():
                update_table(table)

    else:
        # remove the schema from the TAP_SCHEMA (if it exists)
        try:
            tap_schema = TapSchema.objects.get(pk=schema.id)
            tap_schema.delete()

            # remove tables and columns
            for tap_table in tap_schema.tables.all():
                tap_table.columns.all().delete()
                tap_table.delete()

        except TapSchema.DoesNotExist:
            pass


def delete_schema(schema):
    '''
    Remove the schema from the TAP_SCHEMA (if it exists).
    '''
    try:
        TapSchema.objects.get(pk=schema.id).delete()
    except TapSchema.DoesNotExist:
        pass


def update_table(table):
    '''
    Update or create the table in the TAP_SCHEMA.
    '''

    # get the schema from the TAP_SCHEMA
    try:
        tap_schema = TapSchema.objects.get(pk=table.schema.id)
    except TapSchema.DoesNotExist:
        tap_schema = None

    if check_tap_visibility(table) and tap_schema:
        tap_table, created = TapTable.objects.get_or_create(pk=table.id, defaults={'schema': tap_schema})

        tap_table.schema_name = str(table.schema)
        tap_table.table_name = str(table.schema) + '.' + str(table.name)
        tap_table.table_type = table.type
        tap_table.utype = table.utype
        if table.description:
            tap_table.description = table.description[:255]
        tap_table.table_index = table.order
        tap_table.save()

        if created:
            # call the handler for each table of the schema
            for column in table.columns.all():
                update_column(column)

    else:
        # remove the table from the TAP_SCHEMA (if it exists)
        try:
            tap_table = TapTable.objects.get(pk=table.id)
            tap_table.delete()

            # remove columns
            tap_table.columns.all().delete()
        except TapTable.DoesNotExist:
            pass


def delete_table(table):
    '''
    Remove the table from the TAP_SCHEMA (if it exists).
    '''
    try:
        TapTable.objects.get(pk=table.id).delete()
    except TapTable.DoesNotExist:
        pass


def update_column(column):
    '''
    Update or create the column in the TAP_SCHEMA.
    '''
    if settings.METADATA_COLUMN_PERMISSIONS:
        tap_visibility = check_tap_visibility(column)
    else:
        tap_visibility = check_tap_visibility(column.table)

    # get the table from the TAP_SCHEMA
    try:
        tap_table = TapTable.objects.get(pk=column.table.id)
    except TapTable.DoesNotExist:
        tap_table = None

    if tap_visibility and tap_table:
        try:
            tap_column = TapColumn.objects.get(pk=column.id)
            tap_column.table = tap_table
        except TapColumn.DoesNotExist:
            tap_column = TapColumn.objects.create(pk=column.id, table=tap_table)

        tap_column.table_name = str(column.table)
        tap_column.column_name = column.name
        tap_column.datatype = column.datatype
        tap_column.arraysize = column.arraysize
        tap_column.size = column.arraysize
        if column.description:
            tap_column.description = column.description[:255]
        tap_column.utype = column.utype
        tap_column.unit = column.unit
        tap_column.ucd = column.ucd
        tap_column.indexed = column.indexed
        tap_column.principal = column.principal
        tap_column.std = column.std
        tap_column.column_index = column.order
        tap_column.save()
    else:
        # remove the column from the TAP_SCHEMA (if it exists)
        try:
            TapColumn.objects.get(pk=column.id).delete()
        except TapColumn.DoesNotExist:
            pass


def delete_column(column):
    '''
    Remove the column from the TAP_SCHEMA (if it exists).
    '''
    try:
        TapColumn.objects.get(pk=column.id).delete()
    except TapColumn.DoesNotExist:
        pass



from django.contrib import admin

from .models import Profile


class ProfileAdmin(admin.ModelAdmin):
    search_fields = ('user__username', 'user__first_name', 'user__last_name', 'user__email')
    list_display = ('__str__', 'is_pending', 'is_confirmed')
    list_filter = ('is_pending', 'is_confirmed', 'user__is_staff', 'user__is_superuser')
    readonly_fields = ('user', 'consent')


admin.site.register(Profile, ProfileAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-07-07 08:36
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_metadata', '0002_published_for'),
    ]

    operations = [
        migrations.RenameField(
            model_name='column',
            old_name='published_for',
            new_name='groups',
        ),
        migrations.RenameField(
            model_name='database',
            old_name='published_for',
            new_name='groups',
        ),
        migrations.RenameField(
            model_name='function',
            old_name='published_for',
            new_name='groups',
        ),
        migrations.RenameField(
            model_name='table',
            old_name='published_for',
            new_name='groups',
        ),
    ]



from daiquiri.core.renderers import XMLRenderer
from daiquiri.core.utils import get_doi_url


class AvailabilityRenderer(XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):
        self.start('vosi:availability', {
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:vs': 'http://www.ivoa.net/xml/VODataService/v1.0',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VOSI/v1.0 http://www.ivoa.net/xml/VODataService/v1.0 http://www.ivoa.net/xml/VODataService/v1.0'  # noqa: E501
        })
        self.node('available', {}, data.get('available'))
        self.node('note', {}, data.get('note'))
        self.end('vosi:availability')


class CapabilitiesRendererMixin:

    def render_capability(self, capability):
        self.start('capability', {
            'standardID': capability.get('id'),
            'xsi:type': capability.get('type')
        })

        interface = capability.get('interface')
        if interface:
            self.start('interface', {
                'role': interface.get('role'),
                'version': interface.get('version'),
                'xsi:type': interface.get('type')
            })
            access_url = interface.get('access_url')
            if access_url:
                self.node('accessURL', {
                    'use': access_url.get('use'),
                }, access_url.get('url'))

            query_types = interface.get('query_types')
            if query_types:
                for query_type in query_types:
                    self.node('queryType', {}, query_type)

            if interface.get('result_type'):
                self.node('resultType', {}, interface.get('result_type'))

            params = interface.get('params', [])
            for param in params:
                self.start('param', {
                    'std': param.get('std'),
                    'use': param.get('use')
                })
                self.node('name', {}, param.get('name'))
                self.node('description', {}, param.get('description'))

                if param.get('unit'):
                    self.node('unit', {}, param.get('unit'))
                if param.get('ucd'):
                    self.node('ucd', {}, param.get('ucd'))
                self.node('dataType', {}, param.get('datatype'))
                self.end('param')
            self.end('interface')

        for language in capability.get('languages', []):
            self.start('language')
            self.node('name', {}, language.get('name'))
            self.node('version', {}, language.get('version'))
            self.node('description', {}, language.get('description'))
            self.end('language')

        if capability.get('max_sr'):
            self.node('maxSR', {}, capability.get('max_sr'))

        if capability.get('max_records'):
            self.node('maxRecords', {}, capability.get('max_records'))

        if capability.get('verbosity'):
            self.node('verbosity', {}, capability.get('verbosity'))

        for upload_method in capability.get('upload_methods', []):
            self.node('uploadMethod', {'ivo-id': upload_method})

        if capability.get('upload_limit'):
            self.start('uploadLimit')
            self.node('hard', {'unit': 'byte'}, capability.get('upload_limit'))
            self.end('uploadLimit')

        test_query = capability.get('test_query')
        if test_query:
            self.start('testQuery')
            self.node('ra', {}, test_query.get('ra'))
            self.node('dec', {}, test_query.get('dec'))
            self.node('sr', {}, test_query.get('sr'))
            self.end('testQuery')

        self.end('capability')


class CapabilitiesRenderer(CapabilitiesRendererMixin, XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):
        ns = {
            'xmlns:cap': 'http://www.ivoa.net/xml/VOSICapabilities/v1.0',
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:tr': 'http://www.ivoa.net/xml/TAPRegExt/v1.0',
            'xmlns:vr': 'http://www.ivoa.net/xml/VOResource/v1.0',
            'xmlns:vs': 'http://www.ivoa.net/xml/VODataService/v1.1',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        }
        ns['xsi:schemaLocation'] = ' '.join(ns.values())

        self.start('vosi:capabilities', ns)

        for capability in data:
            self.render_capability(capability)

        self.end('vosi:capabilities')


class TablesetRendererMixin:

    def render_tableset(self, tableset, strict=False):
        for schema in tableset:
            self.start('schema')
            self.node('name', {}, schema.get('name'))
            self.node('description', {}, schema.get('description') or '')

            for table in schema.get('tables'):
                if strict:
                    self.start('table')
                else:
                    self.start('table', {
                        'doi': get_doi_url(table.get('doi')),
                        'size': str(table.get('nrows'))
                    })
                self.node('name', {}, table.get('name'))
                self.node('description', {}, table.get('description') or '')

                for column in table['columns']:
                    self.start('column', {'std': 'true'} if column['std'] else {})
                    self.node('name', {}, column['name'])
                    self.node('description', {}, column.get('description') or '')
                    self.node('unit', {}, column.get('unit') or '')
                    self.node('ucd', {}, column.get('ucd') or '')

                    self.render_datatype(column.get('datatype'))
                    for key in ['indexed', 'principal']:
                        if key in column and column.get('key'):
                            self.node('flag', {}, key)

                    self.end('column')

                self.end('table')

            self.end('schema')

    def render_datatype(self, datatype):
        if datatype in ['boolean', 'bit', 'unsignedByte', 'short', 'int', 'long', 'char', 'unicodeChar',
                        'float', 'double', 'floatComplex', 'doubleComplex']:
            self.node('dataType', {'xsi:type': 'vs:VOTableType'}, datatype)
        elif datatype == 'timestamp':
            self.node('dataType', {'xsi:type': 'vs:VOTableType', 'extendedType': 'timestamp'}, 'char')
        else:
            self.node('dataType', {'xsi:type': 'vs:VOTableType'}, 'char')


class TablesetRenderer(TablesetRendererMixin, XMLRenderer):

    def render_document(self, data, accepted_media_type=None, renderer_context=None):
        self.start('vosi:tableset', {
            'xmlns:vosi': 'http://www.ivoa.net/xml/VOSITables/v1.0',
            'xmlns:vs': 'http://www.ivoa.net/xml/VODataService/v1.1',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.ivoa.net/xml/VODataService/v1.1 http://www.ivoa.net/xml/VOSITables/v1.0'
        })
        self.render_tableset(data)
        self.end('vosi:tableset')



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-31 08:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('daiquiri_jobs', '0001_job_model'),
    ]

    operations = [
        migrations.AlterField(
            model_name='job',
            name='execution_duration',
            field=models.PositiveIntegerField(blank=True, default=0),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-14 14:49
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time', models.DateTimeField()),
                ('resource_type', models.CharField(max_length=32)),
                ('resource', jsonfield.fields.JSONField()),
                ('client_ip', models.GenericIPAddressField(blank=True, null=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-time',),
                'verbose_name': 'Record',
                'verbose_name_plural': 'Records',
                'permissions': (('view_record', 'Can view Record'),),
            },
        ),
    ]
